import asyncio
import datetime

import pytest

from chancy import Job, Queue, Chancy, Worker
from chancy.plugins.leadership import ImmediateLeadership
from chancy.plugins.reprioritize import Reprioritize
from chancy.rule import JobRules


def simple_job():
    pass


@pytest.mark.parametrize(
    "chancy",
    [
        {
            "plugins": [
                ImmediateLeadership(),
                Reprioritize(
                    JobRules.Age() > 0,
                    check_interval=1,
                    priority_increase=5,
                ),
            ]
        }
    ],
    indirect=True,
)
@pytest.mark.asyncio
async def test_basic_reprioritization(
    chancy: Chancy, worker: tuple[Worker, asyncio.Task]
):
    """
    Tests that basic reprioritization works as expected.
    """
    await chancy.declare(Queue("default"))

    ref = await chancy.push(
        Job.from_func(simple_job).with_scheduled_at(
            datetime.datetime.now(tz=datetime.timezone.utc)
            + datetime.timedelta(minutes=10)
        )
    )

    initial_job = await chancy.get_job(ref)
    initial_priority = initial_job.priority

    await asyncio.sleep(2)

    updated_job = await chancy.get_job(ref)

    # Priority should have decreased (making it more important)
    assert updated_job.priority > initial_priority



import asyncio
import pytest

from chancy import Job, Queue, QueuedJob, Worker
from chancy.plugins.retry import RetryPlugin


def job_that_fails():
    raise ValueError("This job should fail")


def successful_job():
    pass


@pytest.mark.parametrize(
    "chancy", [{"plugins": [RetryPlugin()]}], indirect=True
)
@pytest.mark.asyncio
async def test_retry_with_default_settings(
    chancy, worker: tuple[Worker, asyncio.Task]
):
    """Test that jobs retry with default settings when no retry_settings in meta"""
    await chancy.declare(Queue("default"))

    ref = await chancy.push(Job.from_func(job_that_fails, max_attempts=3))

    job = await chancy.wait_for_job(ref, timeout=30)
    assert job.state == QueuedJob.State.FAILED
    assert job.attempts == 3
    assert len(job.errors) == 3

    # Verify error structure
    for i, error in enumerate(job.errors):
        assert error["attempt"] == i
        assert "ValueError: This job should fail" in error["traceback"]


@pytest.mark.parametrize(
    "chancy", [{"plugins": [RetryPlugin()]}], indirect=True
)
@pytest.mark.asyncio
async def test_retry_with_custom_settings(
    chancy, worker: tuple[Worker, asyncio.Task]
):
    """Test that jobs retry with custom settings from meta"""
    await chancy.declare(Queue("default"))

    ref = await chancy.push(
        Job.from_func(
            job_that_fails,
            max_attempts=3,
            meta={
                "retry_settings": {
                    "backoff": 2,
                    "backoff_factor": 3,
                    "backoff_limit": 10,
                    "backoff_jitter": [0, 1],
                }
            },
        )
    )

    job = await chancy.wait_for_job(ref, timeout=30)
    assert job.state == QueuedJob.State.FAILED
    assert job.attempts == 3
    assert len(job.errors) == 3

    # Verify the retry settings remained unchanged
    assert job.meta["retry_settings"] == {
        "backoff": 2,
        "backoff_factor": 3,
        "backoff_limit": 10,
        "backoff_jitter": [0, 1],
    }


@pytest.mark.parametrize(
    "chancy", [{"plugins": [RetryPlugin()]}], indirect=True
)
@pytest.mark.asyncio
async def test_no_retry_on_success(chancy, worker: tuple[Worker, asyncio.Task]):
    """Test that successful jobs don't trigger retry logic"""
    await chancy.declare(Queue("default"))

    ref = await chancy.push(
        Job.from_func(
            successful_job,
            max_attempts=3,
            meta={"retry_settings": {"backoff": 1}},
        )
    )

    job = await chancy.wait_for_job(ref, timeout=30)
    assert job.state == QueuedJob.State.SUCCEEDED
    assert job.attempts == 1
    assert len(job.errors) == 0


@pytest.mark.parametrize(
    "chancy", [{"plugins": [RetryPlugin()]}], indirect=True
)
@pytest.mark.asyncio
async def test_respect_max_attempts(
    chancy, worker: tuple[Worker, asyncio.Task]
):
    """Test that jobs don't retry beyond max_attempts"""
    await chancy.declare(Queue("default"))

    ref = await chancy.push(
        Job.from_func(
            job_that_fails,
            max_attempts=2,
            meta={
                "retry_settings": {
                    "backoff": 1,
                    "backoff_factor": 2,
                    "backoff_limit": 10,
                    "backoff_jitter": [0, 1],
                }
            },
        )
    )

    job = await chancy.wait_for_job(ref, timeout=30)
    assert job.state == QueuedJob.State.FAILED
    assert job.attempts == 2
    assert len(job.errors) == 2



import asyncio
import secrets
from typing import AsyncIterator

import pytest
import pytest_asyncio
import sys
from pytest_postgresql import factories

from chancy import Chancy, Worker


external_postgres = factories.postgresql_noproc(
    host="localhost",
    password="localtest",
    user="postgres",
    port=8190,
    dbname=f"chancy_test_{secrets.token_hex(8)}",
)
postgresql = factories.postgresql(
    "external_postgres",
)


@pytest_asyncio.fixture()
async def chancy(request, postgresql):
    """
    Provides a Chancy application instance with an open connection pool
    to the test database.
    """
    i = postgresql.info

    async with Chancy(
        dsn=f"postgresql://{i.user}:{i.password}@{i.host}:{i.port}/{i.dbname}",
        **getattr(request, "param", {}),
    ) as chancy:
        await chancy.migrate()
        yield chancy
        await chancy.migrate(to_version=0)


@pytest.fixture
def chancy_just_app(postgresql):
    """
    Provides just a configured chancy instance with no open connection pool
    or migrations.
    """
    i = postgresql.info
    return Chancy(
        dsn=f"postgresql://{i.user}:{i.password}@{i.host}:{i.port}/{i.dbname}",
    )


@pytest_asyncio.fixture()
async def worker(request, chancy) -> AsyncIterator[tuple[Worker, asyncio.Task]]:
    """
    Starts and returns a Worker and the task associated with it.

    If the worker is not stopped by the time the test completes, it will be
    cancelled.
    """
    async with Worker(chancy, **getattr(request, "param", {})) as worker:
        yield worker


@pytest_asyncio.fixture()
async def worker_no_start(chancy) -> Worker:
    """
    Returns a Worker instance that has not been started.
    """
    return Worker(chancy)


@pytest.fixture(
    params=(
        [
            "chancy.executors.process.ProcessExecutor",
            "chancy.executors.thread.ThreadedExecutor",
        ]
        + ["chancy.executors.sub.SubInterpreterExecutor"]
        if sys.version_info >= (3, 13)
        else []
    )
)
def sync_executor(request):
    """
    Provides a parameterized fixture for all sync executors.
    """
    return request.param


@pytest.fixture(
    params=[
        "chancy.executors.asyncex.AsyncExecutor",
    ]
)
def async_executor(request):
    """
    Provides a parameterized fixture for all async executors.
    """
    return request.param



from psycopg import AsyncCursor
from psycopg import sql

from chancy.app import Chancy
from chancy.worker import Worker
from chancy.utils import timed_block
from chancy.plugin import Plugin, PluginScope


class Recovery(Plugin):
    """
    Recovers jobs that appear to be abandoned by a worker.

    Typically, this happens when a worker is unexpectedly terminated, or has
    otherwise been lost which we recognize by checking the last seen timestamp
    of the worker heartbeat.

    This will transition any matching jobs back to the "pending" state, and
    increment the `max_attempts` counter by 1 to allow it to be retried.

    :param poll_interval: The number of seconds between recovery poll intervals.
    """

    def __init__(self, *, poll_interval: int = 60):
        super().__init__()
        self.poll_interval = poll_interval

    @classmethod
    def get_scope(cls) -> PluginScope:
        return PluginScope.WORKER

    async def run(self, worker: Worker, chancy: Chancy):
        while await self.sleep(self.poll_interval):
            await self.wait_for_leader(worker)
            async with chancy.pool.connection() as conn:
                async with conn.cursor() as cursor:
                    with timed_block() as chancy_time:
                        rows_recovered = await self.recover(
                            worker, chancy, cursor
                        )
                        chancy.log.info(
                            f"Recovery recovered {rows_recovered} row(s) from"
                            f" the database. Took {chancy_time.elapsed:.2f}"
                            f" seconds."
                        )
                        await chancy.notify(
                            cursor,
                            "recovery.recovered",
                            {
                                "elapsed": chancy_time.elapsed,
                                "rows_recovered": rows_recovered,
                            },
                        )

    @classmethod
    async def recover(
        cls, worker: Worker, chancy: Chancy, cursor: AsyncCursor
    ) -> int:
        """
        Recover jobs that were running when the worker was unexpectedly
        terminated, or has otherwise been lost.

        :param worker: The worker that is running the recovery.
        :param chancy: The Chancy application.
        :param cursor: The cursor to use for database operations.
        :return: The number of rows recovered from the database
        """
        query = sql.SQL(
            """
            UPDATE
                {jobs} cj
            SET
                state = 'pending',
                taken_by = NULL,
                started_at = NULL,
                max_attempts = max_attempts + 1
            WHERE
               NOT EXISTS (
                    SELECT 1
                    FROM {workers} cw
                    WHERE (
                        cw.worker_id = cj.taken_by
                        AND
                        cw.last_seen >= NOW() - INTERVAL '{interval} SECOND'
                    )
              )
              AND state = 'running';
        """
        ).format(
            jobs=sql.Identifier(f"{chancy.prefix}jobs"),
            workers=sql.Identifier(f"{chancy.prefix}workers"),
            interval=sql.Literal(worker.heartbeat_timeout),
        )

        await cursor.execute(query)
        return cursor.rowcount



from psycopg import AsyncCursor
from psycopg import sql

from chancy.migrate import Migration, Migrator


class V1Migration(Migration):
    async def up(self, migrator: Migrator, cursor: AsyncCursor):
        """
        Create the initial $prefix_jobs, $prefix_leaders tables.
        """
        await cursor.execute(
            sql.SQL(
                """
                CREATE TABLE {jobs} (
                    id UUID PRIMARY KEY,
                    queue TEXT NOT NULL,
                    func TEXT NOT NULL,
                    kwargs JSON NOT NULL DEFAULT '{{}}',
                    limits JSON NOT NULL DEFAULT '[]',
                    meta JSON NOT NULL DEFAULT '{{}}',
                    state VARCHAR (25) NOT NULL DEFAULT 'pending',
                    priority INTEGER DEFAULT 0,
                    attempts INTEGER DEFAULT 0,
                    max_attempts INTEGER DEFAULT 1,
                    taken_by TEXT,
                    created_at TIMESTAMPTZ NOT NULL DEFAULT NOW(),
                    started_at TIMESTAMPTZ,
                    completed_at TIMESTAMPTZ,
                    scheduled_at TIMESTAMPTZ NOT NULL DEFAULT NOW(),
                    unique_key TEXT,
                    errors JSON NOT NULL DEFAULT '[]'
                )
                """
            ).format(jobs=sql.Identifier(f"{migrator.prefix}jobs"))
        )
        await cursor.execute(
            sql.SQL(
                """
                CREATE UNIQUE INDEX {jobs_unique_key}
                ON {jobs} (unique_key)
                WHERE
                    unique_key IS NOT NULL
                        AND state NOT IN ('succeeded', 'failed');
                """
            ).format(
                jobs_unique_key=sql.Identifier(
                    f"{migrator.prefix}jobs_unique_key"
                ),
                jobs=sql.Identifier(f"{migrator.prefix}jobs"),
            )
        )
        await cursor.execute(
            sql.SQL(
                """
                CREATE UNLOGGED TABLE {leader} (
                    id BIGSERIAL PRIMARY KEY,
                    worker_id TEXT NOT NULL,
                    first_seen TIMESTAMPTZ NOT NULL DEFAULT NOW(),
                    expires_at TIMESTAMPTZ NOT NULL
                );
                ALTER TABLE {leader}
                    ADD CONSTRAINT leader_worker_id_unique UNIQUE
                        (worker_id);
                """
            ).format(leader=sql.Identifier(f"{migrator.prefix}leader"))
        )

        await cursor.execute(
            sql.SQL(
                """
                CREATE UNLOGGED TABLE {workers} (
                    -- The UUID of the worker.
                    worker_id TEXT NOT NULL PRIMARY KEY,
                    -- The tags associated with this worker.
                    tags TEXT[],
                    -- The queues this worker is currently processing.
                    queues TEXT[],
                    last_seen TIMESTAMPTZ NOT NULL DEFAULT NOW(),
                    expires_at TIMESTAMPTZ NOT NULL
                );
                """
            ).format(workers=sql.Identifier(f"{migrator.prefix}workers"))
        )

        await cursor.execute(
            sql.SQL(
                """
                CREATE UNLOGGED TABLE {rate_limit} (
                    queue VARCHAR(255) PRIMARY KEY,
                    window_start INTEGER NOT NULL,
                    count INTEGER NOT NULL DEFAULT 0
                );
                """
            ).format(
                rate_limit=sql.Identifier(f"{migrator.prefix}queue_rate_limits")
            )
        )

        await cursor.execute(
            sql.SQL(
                """
                CREATE TABLE {queues} (
                    -- A unique name for the queue.
                    name TEXT PRIMARY KEY,
                    -- The state of the queue. This can be used to pause
                    -- processing of jobs in the queue.
                    state TEXT NOT NULL DEFAULT 'active',
                    -- The number of jobs that can be processed concurrently
                    -- per worker.
                    concurrency INTEGER DEFAULT 1,
                    -- Only worker's matching these tags will be able to
                    -- process jobs from this queue, or None for all workers.
                    tags TEXT[],
                    -- The executor to use for processing jobs in this queue.
                    executor TEXT NOT NULL DEFAULT
                        'chancy.executors.process.ProcessExecutor',
                    -- Options to pass to the executor.
                    executor_options JSON,
                    -- The interval at which the queue should be polled for
                    -- jobs.
                    polling_interval INTEGER DEFAULT 1,
                    -- The time at which the queue was first declared.
                    created_at TIMESTAMPTZ NOT NULL DEFAULT NOW(),
                    -- The number of jobs that can be processed per period.
                    rate_limit INTEGER,
                    -- The period of time over which the rate limit applies.
                    rate_limit_window INTEGER
                );
                """
            ).format(queues=sql.Identifier(f"{migrator.prefix}queues"))
        )

    async def down(self, migrator: Migrator, cursor: AsyncCursor):
        """
        Drop the $prefix_jobs, $prefix_leaders tables.
        """
        await cursor.execute(
            sql.SQL("DROP TABLE IF EXISTS {jobs}").format(
                jobs=sql.Identifier(f"{migrator.prefix}jobs")
            )
        )
        await cursor.execute(
            sql.SQL("DROP TABLE IF EXISTS {leader}").format(
                leader=sql.Identifier(f"{migrator.prefix}leader")
            )
        )
        await cursor.execute(
            sql.SQL("DROP TABLE IF EXISTS {workers}").format(
                workers=sql.Identifier(f"{migrator.prefix}workers")
            )
        )
        await cursor.execute(
            sql.SQL("DROP TABLE IF EXISTS {queues}").format(
                queues=sql.Identifier(f"{migrator.prefix}queues")
            )
        )
        await cursor.execute(
            sql.SQL("DROP TABLE IF EXISTS {rate_limit}").format(
                rate_limit=sql.Identifier(f"{migrator.prefix}queue_rate_limits")
            )
        )



import dataclasses
from datetime import datetime, timedelta, timezone
import random

from chancy.plugin import Plugin, PluginScope
from chancy.job import QueuedJob
from chancy.worker import Worker


class RetryPlugin(Plugin):
    """
    Plugin that handles job retries based on settings stored in the Job's
    metadata.

    This plugin can be used as an example for implementing your own retry
    policies.

    Usage:

    .. code-block:: python

        def job_that_fails():
            raise ValueError("This job should fail.")

        async with Chancy(..., plugins=[
            RetryPlugin()
        ]) as chancy:
            await chancy.declare(Queue("default"))
            await chancy.push(
                Job.from_func(
                    job_that_fails,
                    max_attempts=3
                ).with_meta({
                    "retry_settings": {
                        "backoff": 2,
                        "backoff_factor": 3,
                        "backoff_limit": 10,
                        "backoff_jitter": [0, 1],
                    }
                })
            )

    """

    @classmethod
    def get_scope(cls) -> PluginScope:
        return PluginScope.WORKER

    @staticmethod
    def calculate_next_run(job: QueuedJob, retry_settings: dict) -> datetime:
        delay = retry_settings.get("backoff", 1)
        factor = retry_settings.get("backoff_factor", 2.0)
        limit = retry_settings.get("backoff_limit", 300)
        jitter = retry_settings.get("backoff_jitter", [1, 5])

        delay *= factor ** (job.attempts - 1)
        delay = min(delay, limit)
        delay += random.uniform(*jitter)

        return datetime.now(timezone.utc) + timedelta(seconds=delay)

    async def on_job_completed(
        self, job: QueuedJob, worker: Worker, *, exc: Exception | None = None
    ) -> QueuedJob:
        if exc is None or job.state not in {
            job.State.FAILED,
            job.State.RETRYING,
        }:
            return job

        if job.attempts >= job.max_attempts:
            return job

        retry_settings = job.meta.get("retry_settings", {})

        # We don't need to adjust the # of attempts as the base executor will
        # do that for us.
        return dataclasses.replace(
            job,
            state=QueuedJob.State.RETRYING,
            scheduled_at=self.calculate_next_run(job, retry_settings),
            completed_at=None,
        )



from dataclasses import asdict

from psycopg import sql
from psycopg.rows import dict_row
from starlette.responses import Response

from chancy.plugins.api import ApiPlugin
from chancy.plugins.workflow import WorkflowPlugin
from chancy.utils import json_dumps


class WorkflowApiPlugin(ApiPlugin):
    """
    API plugin for workflows.
    """

    def name(self):
        return "workflow"

    def routes(self):
        return [
            {
                "path": "/api/v1/workflows",
                "endpoint": self.get_workflows,
                "methods": ["GET"],
                "name": "get_workflows",
            },
            {
                "path": "/api/v1/workflows/{id}",
                "endpoint": self.get_workflow,
                "methods": ["GET"],
                "name": "get_workflow",
            },
        ]

    @staticmethod
    async def get_workflows(request, *, chancy, worker):
        """
        Get all known workflows.
        """
        async with chancy.pool.connection() as conn:
            async with conn.cursor(row_factory=dict_row) as cursor:
                await cursor.execute(
                    sql.SQL(
                        """
                        SELECT
                            w.*
                        FROM
                            {workflows_table} w
                        ORDER BY
                            created_at DESC;
                        """
                    ).format(
                        workflows_table=sql.Identifier(
                            f"{chancy.prefix}workflows"
                        ),
                    )
                )

                results = await cursor.fetchall()

                return Response(
                    json_dumps(results),
                    media_type="application/json",
                )

    @staticmethod
    async def get_workflow(request, *, chancy, worker):
        """
        Get a single workflow by ID.
        """
        workflow_id = request.path_params["id"]
        workflow = await WorkflowPlugin.fetch_workflow(chancy, workflow_id)

        return Response(
            json_dumps(asdict(workflow)),
            media_type="application/json",
        )



import dataclasses
import enum
from dataclasses import KW_ONLY


@dataclasses.dataclass(frozen=True)
class Queue:
    """
    Represents a queue in a Chancy cluster.

    Queues are used to group jobs together and determine how they should be
    processed. Each queue has a name, concurrency level, and a set of tags that
    determine which workers can process jobs from the queue.

    Queue's must be declared using :func:`~chancy.app.Chancy.declare` before workers
    will be able to process jobs from them.

    .. code-block:: python

        async with Chancy(dsn="postgresql://localhost/postgres") as chancy:
            chancy.declare(Queue(name="default", concurrency=4))

    By default, this queue will shortly be picked up by all running workers and
    begin processing jobs. If you want to instead apply it to specific workers,
    you can assign it using "tags":

    .. code-block:: python

        async with Chancy(dsn="postgresql://localhost/postgres") as chancy:
            chancy.declare(Queue(name="default", concurrency=4, tags={"reporting"}))

    This will only be picked up by workers that have the "reporting" tag:

    .. code-block:: python

        async with Chancy(dsn="postgresql://localhost/postgres") as chancy:
            await Worker(chancy, tags={"reporting"}).start()

    Queues can have a global rate limit applied to them, which will be enforced
    across all workers processing jobs from the queue:

    .. code-block:: python

        async with Chancy(dsn="postgresql://localhost/postgres") as chancy:
            chancy.declare(
                Queue(name="default", rate_limit=10, rate_limit_window=60)
            )

    This will limit the queue to processing 10 jobs per minute across all
    workers. If the rate limit is exceeded, jobs will be skipped until the
    rate limit window has passed. Combined with the AsyncExecutor, this can be
    a very easy way to work with external APIs.

    .. note::

        Rate limiting is done with a fixed window algorithm for simplicity.
        If you need to do something custom, subclass the worker and
        re-implement :func:`~chancy.worker.Worker.fetch_jobs`.
    """

    class State(enum.Enum):
        #: The queue is active and jobs can be processed.
        ACTIVE = "active"
        #: The queue is paused and no jobs will be processed.
        PAUSED = "paused"

    #: A globally unique identifier for the queue.
    name: str

    _ = KW_ONLY
    #: The number of jobs that can be processed concurrently per worker.
    concurrency: int = 1
    #: The tags that determine which workers will process this queue.
    tags: set[str] = dataclasses.field(default_factory=lambda: {r".*"})
    #: The state of the queue.
    state: State = State.ACTIVE
    #: The import path to the executor that should be used to process jobs in
    #: this queue.
    executor: str = "chancy.executors.process.ProcessExecutor"
    #: The options to pass to the executor's constructor.
    executor_options: dict = dataclasses.field(default_factory=dict)
    #: The number of seconds to wait between polling the queue for new jobs.
    polling_interval: int = 1
    #: An optional global rate limit to apply to this queue. All workers
    #: processing jobs from this queue will be subject to this limit.
    rate_limit: int | None = None
    #: The period of time over which the rate limit applies (in seconds).
    rate_limit_window: int | None = None

    @classmethod
    def unpack(cls, data: dict) -> "Queue":
        """
        Unpack a serialized queue object into a Queue instance.
        """
        return cls(
            name=data["name"],
            concurrency=data["concurrency"],
            tags=set(data["tags"]),
            state=cls.State(data["state"]),
            executor=data["executor"],
            executor_options=data["executor_options"],
            polling_interval=data["polling_interval"],
            rate_limit=data.get("rate_limit"),
            rate_limit_window=data.get("rate_limit_window"),
        )

    def pack(self) -> dict:
        """
        Pack the queue into a dictionary that can be serialized and used to
        recreate the queue later.
        """
        return {
            "name": self.name,
            "concurrency": self.concurrency,
            "tags": list(self.tags),
            "state": self.state.value,
            "executor": self.executor,
            "executor_options": self.executor_options,
            "polling_interval": self.polling_interval,
            "rate_limit": self.rate_limit,
            "rate_limit_window": self.rate_limit_window,
        }



from psycopg import AsyncCursor
from psycopg import sql

from chancy.migrate import Migration, Migrator


class V1Migration(Migration):
    async def up(self, migrator: Migrator, cursor: AsyncCursor):
        await cursor.execute(
            sql.SQL(
                """
                CREATE TABLE {table} (
                    unique_key TEXT PRIMARY KEY,
                    job JSON NOT NULL,
                    cron TEXT NOT NULL,
                    last_run TIMESTAMPTZ,
                    next_run TIMESTAMPTZ NOT NULL
                )
                """
            ).format(table=sql.Identifier(f"{migrator.prefix}cron"))
        )

    async def down(self, migrator: Migrator, cursor: AsyncCursor):
        await cursor.execute(
            sql.SQL("DROP TABLE {table}").format(
                table=sql.Identifier(f"{migrator.prefix}cron")
            )
        )



import asyncio
from functools import wraps


def run_async_command(f):
    """
    Decorator to run a coroutine as a Click command.
    """

    @wraps(f)
    def wrapper(*args, **kwargs):
        return asyncio.run(f(*args, **kwargs))

    return wrapper



import abc
import inspect
import typing
import traceback
import dataclasses
from datetime import datetime, timezone
from typing import get_type_hints, Callable

from chancy.job import QueuedJob
from chancy.queue import Queue

if typing.TYPE_CHECKING:
    from chancy.worker import Worker


class Executor(abc.ABC):
    """
    The base class for all executors.

    Executors are responsible for managing the execution of jobs after they've
    been retrieved from a queue.

    See :class:`~chancy.executors.process.ProcessExecutor` and
    :class:`~chancy.executors.asyncex.AsyncExecutor` for examples of built-in
    executors.
    """

    def __init__(self, worker: "Worker", queue: "Queue"):
        self.worker = worker
        self.queue = queue

    @abc.abstractmethod
    async def push(self, job: QueuedJob):
        """
        Push a job onto the job pool.
        """

    async def job_completed(self, job: QueuedJob, exc: Exception | None = None):
        """
        Called when a job has completed.

        This method should be called by the executor when a job has completed
        execution. It will update the job's state in the queue and handle
        retries if necessary.

        :param job: The job that has completed.
        :param exc: The exception that was raised during execution, if any.
        """
        if exc is None:
            now = datetime.now(tz=timezone.utc)
            new_instance = dataclasses.replace(
                job,
                state=QueuedJob.State.SUCCEEDED,
                completed_at=now,
                attempts=job.attempts + 1,
            )
        else:
            is_failure = job.attempts + 1 >= job.max_attempts

            new_state = (
                QueuedJob.State.FAILED
                if is_failure
                else QueuedJob.State.RETRYING
            )

            new_instance = dataclasses.replace(
                job,
                state=new_state,
                attempts=job.attempts + 1,
                completed_at=(
                    datetime.now(tz=timezone.utc) if is_failure else None
                ),
                errors=[
                    *job.errors,
                    {
                        "traceback": "".join(
                            traceback.format_exception(
                                type(exc), exc, exc.__traceback__
                            )
                        ),
                        "attempt": job.attempts,
                    },
                ],
            )

        # Each plugin has a chance to modify the job instance after it's
        # completed.
        for plugin in self.worker.chancy.plugins:
            try:
                new_instance = await plugin.on_job_completed(
                    new_instance, worker=self.worker, exc=exc
                )
            except NotImplementedError:
                continue

        await self.worker.queue_update(new_instance)

    @staticmethod
    def get_function_and_kwargs(job: QueuedJob) -> tuple[Callable, dict]:
        """
        Finds the function which should be executed for the given job and
        returns its keyword arguments.

        :param job: The job instance to get the function and arguments for.
        :return: A tuple containing the function and its keyword arguments.
        """
        mod_name, func_name = job.func.rsplit(".", 1)
        mod = __import__(mod_name, fromlist=[func_name])
        try:
            func = getattr(mod, func_name)
        except AttributeError:
            raise AttributeError(
                f"Could not find function {func_name} in module {mod_name}."
            )

        # We take a look at the type signature for the function to see if the
        # user has specified that the job instance should be passed as a
        # keyword argument.
        type_hints = get_type_hints(func)
        sig = inspect.signature(func)
        kwargs = job.kwargs or {}
        for param_name, param in sig.parameters.items():
            if (
                param.kind == param.KEYWORD_ONLY
                and type_hints.get(param_name) is QueuedJob
            ):
                kwargs[param_name] = job

        return func, kwargs

    async def stop(self):
        """
        Stop the executor.
        """

    @abc.abstractmethod
    def __len__(self):
        """
        Get the number of pending and running jobs in the pool.
        """

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.stop()



import enum
from dataclasses import dataclass, field
from datetime import datetime
from typing import List, Dict, TextIO, Self
from psycopg import sql
from psycopg.rows import dict_row

from chancy.plugin import Plugin, PluginScope
from chancy.app import Chancy
from chancy.worker import Worker
from chancy.job import Job, QueuedJob
from chancy.utils import json_dumps, chancy_uuid
from chancy.rule import Rule


@dataclass
class WorkflowStep:
    """
    A single step in a workflow.
    """

    #: The job to execute when this step is ready.
    job: Job
    #: The unique ID of the step.
    step_id: str
    #: A list of step IDs that this step depends on.
    dependencies: List[str] = field(default_factory=list)
    #: The current state of the step.
    state: QueuedJob.State | None = QueuedJob.State.PENDING
    #: The unique ID of a running Job which is associated with this step.
    job_id: str | None = None


@dataclass
class Workflow:
    """
    A simple dependency-based workflow.
    """

    class State(enum.Enum):
        PENDING = "pending"
        RUNNING = "running"
        COMPLETED = "completed"
        FAILED = "failed"

    # A descriptive name for the workflow.
    name: str
    #: A dictionary of steps in the workflow, keyed by step ID.
    steps: Dict[str, WorkflowStep] = field(default_factory=dict)
    #: The current state of the workflow.
    state: State = State.PENDING

    #: The unique ID of a specific run of the workflow.
    id: str = field(default_factory=chancy_uuid)
    #: The time the workflow was created.
    created_at: datetime | None = None
    #: The time the workflow was last updated.
    updated_at: datetime | None = None

    def add_step(
        self, step_id: str, job: Job, dependencies: List[str] = None
    ) -> "Workflow":
        """
        Add a step to the workflow.

        :param step_id: The ID of the step.
        :param job: The job to execute.
        :param dependencies: A list of step IDs that this step depends on.
        """
        self.steps[step_id] = WorkflowStep(
            job=job, dependencies=dependencies or [], step_id=step_id
        )
        return self

    def __repr__(self):
        return f"<Workflow({self.name!r}, {self.state!r})>"

    def __len__(self) -> int:
        return len(self.steps)

    def __iadd__(self, other: WorkflowStep) -> "Workflow":
        self.steps[other.step_id] = other
        return self

    def __iter__(self):
        return iter(self.steps.items())

    def __getitem__(self, key: str) -> WorkflowStep:
        return self.steps[key]

    def __delitem__(self, key: str):
        del self.steps[key]

    @property
    def steps_by_state(self) -> Dict[QueuedJob.State, List[WorkflowStep]]:
        steps_by_state = {}
        for step in self.steps.values():
            steps_by_state.setdefault(step.state, []).append(step)
        return steps_by_state


class WorkflowPlugin(Plugin):
    """
    Support for simple dependency-based workflows.

    Workflows are defined by a series of steps, each of which can depend on
    one or more other steps. When all dependencies are met, the step is
    executed. This forms a directed acyclic graph (DAG) of steps that can be
    visualized as a workflow.

    If notifications are enabled (the default) workflows will run almost
    immediately after creation.

    You can use the :func:`generate_dot` method to generate a DOT file
    representation of a workflow, which can be visualized using Graphviz or
    similar tools. This can be a helpful debugging and validation tool for
    complex workflows.

    Enable the plugin by adding it to the list of plugins in the Chancy
    constructor:

    .. code-block:: python

        from chancy.plugins.workflow import WorkflowPlugin

        async with Chancy(
            dsn="postgresql://localhost/postgres",
            plugins=[WorkflowPlugin()]
        ) as chancy:
            ...

    Example
    -------

    .. code-block:: python
       :caption: example_workflow.py

        import asyncio
        from chancy import Job, Chancy
        from chancy.plugins.workflow import (
            Workflow,
            WorkflowStep,
            WorkflowPlugin
        )

        async def top():
            print(f"Top")

        async def left():
            print(f"Left")

        async def right():
            print(f"Right")

        async def bottom():
            print(f"Bottom")

        async def main():
            async with Chancy(dsn="postgresql://localhost/postgres") as chancy:
                workflow = Workflow("example")

                workflow += WorkflowStep(
                    job=Job.from_func(top),
                    step_id="top",
                )
                workflow += WorkflowStep(
                    job=Job.from_func(left),
                    step_id="left",
                    dependencies=["top"],
                )
                workflow += WorkflowStep(
                    job=Job.from_func(right),
                    step_id="right",
                    dependencies=["top"],
                )
                workflow += WorkflowStep(
                    job=Job.from_func(bottom),
                    step_id="bottom",
                    dependencies=["left", "right"],
                )

                await WorkflowPlugin.push(chancy, workflow)

        if __name__ == "__main__":
            asyncio.run(main())

    If we visualize our newly created workflow using :func:`generate_dot`, we
    get:

    .. graphviz:: _static/workflow.dot

    :param polling_interval: The interval at which to poll for new workflows.
    :param max_workflows_per_run: The maximum number of workflows to process
                                  in a single run of the plugin.
    """

    class Rules:
        class Age(Rule):
            def __init__(self):
                super().__init__("age")

            def to_sql(self) -> sql.Composable:
                return sql.SQL("EXTRACT(EPOCH FROM (NOW() - created_at))")

    def __init__(
        self,
        *,
        polling_interval: int = 1,
        max_workflows_per_run: int = 1000,
        pruning_rule: Rule = Rules.Age() > 60 * 60 * 24,
    ):
        super().__init__()
        self.polling_interval = polling_interval
        self.max_workflows_per_run = max_workflows_per_run
        self.pruning_rule = pruning_rule

    @classmethod
    def get_scope(cls) -> PluginScope:
        return PluginScope.WORKER

    async def run(self, worker: Worker, chancy: Chancy):
        worker.hub.on(
            "workflow.created",
            lambda *args, **kwargs: self.wake_up(),
        )

        while await self.sleep(self.polling_interval):
            await self.wait_for_leader(worker)
            workflows = await self.fetch_workflows(
                chancy,
                states=["pending", "running"],
                limit=self.max_workflows_per_run,
            )
            for workflow in workflows:
                await self.process_workflow(chancy, workflow)

    @classmethod
    async def fetch_workflow(cls, chancy: Chancy, id_: str) -> Workflow | None:
        """
        Fetch a single workflow from the database.

        :param chancy: The Chancy application.
        :param id_: The ID of the workflow to fetch.
        :return: The workflow, or None if it does not exist.
        """
        workflows = await cls.fetch_workflows(chancy, ids=[id_])
        return workflows[0] if workflows else None

    @staticmethod
    async def fetch_workflows(
        chancy: Chancy,
        *,
        states: list[str] | None = None,
        ids: list[str] | None = None,
        limit: int = 100,
    ) -> List[Workflow]:
        """
        Fetch workflows from the database, optionally matching the given
        conditions.

        :param chancy: The Chancy application.
        :param states: A list of states to match.
        :param ids: A list of IDs to match.
        :param limit: The maximum number of workflows to fetch.
        :return: A list of workflows.
        """
        async with chancy.pool.connection() as conn:
            async with conn.cursor(row_factory=dict_row) as cursor:
                await cursor.execute(
                    sql.SQL(
                        """
                        SELECT 
                            w.id, 
                            w.name, 
                            w.state,
                            w.created_at,
                            w.updated_at,
                            COALESCE(json_agg(
                                json_build_object(
                                    'step_id', ws.step_id,
                                    'job_data', ws.job_data,
                                    'dependencies', ws.dependencies,
                                    'state', j.state,
                                    'job_id', ws.job_id
                                )
                            ), '[]'::json) as steps
                        FROM {workflows} w
                        LEFT JOIN {workflow_steps} ws ON w.id = ws.workflow_id
                        LEFT JOIN {jobs} j ON ws.job_id = j.id
                        WHERE (
                            %(states)s::text[] IS NULL OR
                            w.state = ANY(%(states)s::text[])
                        ) AND (
                            %(ids)s::uuid[] IS NULL OR
                            w.id = ANY(%(ids)s::uuid[])
                        )
                        GROUP BY w.id, w.name, w.state
                        LIMIT {limit}
                        """
                    ).format(
                        workflows=sql.Identifier(f"{chancy.prefix}workflows"),
                        workflow_steps=sql.Identifier(
                            f"{chancy.prefix}workflow_steps"
                        ),
                        jobs=sql.Identifier(f"{chancy.prefix}jobs"),
                        limit=sql.Literal(limit),
                    ),
                    {
                        "states": states,
                        "ids": ids,
                    },
                )
                rows = await cursor.fetchall()

                return [
                    Workflow(
                        id=row["id"],
                        name=row["name"],
                        state=Workflow.State(row["state"]),
                        updated_at=row["updated_at"],
                        created_at=row["created_at"],
                        steps={
                            step["step_id"]: WorkflowStep(
                                job=Job.unpack(step["job_data"]),
                                dependencies=step["dependencies"],
                                state=(
                                    QueuedJob.State(step["state"])
                                    if step["state"]
                                    else None
                                ),
                                step_id=step["step_id"],
                                job_id=step["job_id"],
                            )
                            for step in row["steps"]
                        },
                    )
                    for row in rows
                ]

    async def process_workflow(self, chancy: Chancy, workflow: Workflow):
        """
        Process a single iteration of the given workflow, progressing the
        state of each step and the overall workflow as necessary.

        :param chancy: The Chancy application.
        :param workflow: The workflow to process.
        """
        # If the workflow is already in a terminal state, there's no further
        # processing to do, although we may add future state handling here
        # for retries.
        if workflow.state in [Workflow.State.COMPLETED, Workflow.State.FAILED]:
            return

        # We check each step in the workflow to see:
        #    - If it has an associated job, and if so, what's the state of it?
        #    - If it has any dependencies, and if so, are they all completed?
        # If all dependencies are met, we can execute the job.
        for step_id, step in workflow.steps.items():
            # If the step is already in a terminal state, we can skip it.
            if step.state in [
                QueuedJob.State.SUCCEEDED,
                QueuedJob.State.FAILED,
            ]:
                continue

            dependencies = [workflow.steps[dep] for dep in step.dependencies]
            if all(
                dep.state == QueuedJob.State.SUCCEEDED for dep in dependencies
            ):
                if step.job_id is None:
                    step.job_id = (await chancy.push(step.job)).identifier

        # Are all jobs complete, or any jobs failed? If so, we can mark the
        # workflow as completed or failed.
        states = workflow.steps_by_state
        if len(states.get(QueuedJob.State.SUCCEEDED, [])) == len(workflow):
            workflow.state = Workflow.State.COMPLETED
        elif states.get(QueuedJob.State.FAILED):
            workflow.state = Workflow.State.FAILED

        # We update the workflow in the database to reflect the new state
        await self.push(chancy, workflow)

    @classmethod
    async def push(cls, chancy: Chancy, workflow: Workflow) -> str:
        """
        Push new workflow to the database.

        If the workflow already exists in the database, it will be updated
        instead.

        :param chancy: The Chancy application.
        :param workflow: The workflow to push.
        :return: The UUID of the newly created workflow.
        """
        async with chancy.pool.connection() as conn:
            async with conn.transaction():
                async with conn.cursor() as cursor:
                    await cursor.execute(
                        sql.SQL(
                            """
                            INSERT INTO {workflows} (
                                id,
                                name,
                                state,
                                created_at,
                                updated_at
                            )
                            VALUES (%s, %s, %s, NOW(), NOW())
                            ON CONFLICT (id) DO UPDATE
                            SET name = EXCLUDED.name,
                                state = EXCLUDED.state,
                                updated_at = NOW()
                            RETURNING id, created_at, updated_at
                            """
                        ).format(
                            workflows=sql.Identifier(
                                f"{chancy.prefix}workflows"
                            )
                        ),
                        [workflow.id, workflow.name, workflow.state.value],
                    )
                    result = await cursor.fetchone()
                    workflow.id, workflow.created_at, workflow.updated_at = (
                        result
                    )

                    for step_id, step in workflow.steps.items():
                        await cursor.execute(
                            sql.SQL(
                                """
                                INSERT INTO {workflow_steps} (
                                    workflow_id,
                                    step_id,
                                    job_data,
                                    dependencies,
                                    job_id
                                )
                                VALUES (%s, %s, %s, %s, %s)
                                ON CONFLICT (workflow_id, step_id) DO UPDATE
                                SET job_data = EXCLUDED.job_data,
                                    dependencies = EXCLUDED.dependencies,
                                    job_id = EXCLUDED.job_id,
                                    updated_at = NOW()
                                """
                            ).format(
                                workflow_steps=sql.Identifier(
                                    f"{chancy.prefix}workflow_steps"
                                )
                            ),
                            [
                                workflow.id,
                                step_id,
                                json_dumps(step.job.pack()),
                                json_dumps(step.dependencies),
                                step.job_id,
                            ],
                        )

                    await chancy.notify(
                        cursor,
                        "workflow.upserted",
                        {
                            "id": workflow.id,
                            "name": workflow.name,
                        },
                    )

        return workflow.id

    @staticmethod
    def generate_dot(workflow: Workflow, output: TextIO):
        """
        Generate a DOT file representation of the workflow.

        :param workflow: The Workflow object to visualize.
        :param output: A file-like object to write the DOT content to.
        """
        # Start the digraph
        output.write(f'digraph "{workflow.name}" {{\n')
        output.write("  rankdir=TB;\n")
        output.write(
            '  node [shape=box, style="rounded,filled", fontname="Arial"];\n'
        )

        # Define color scheme
        colors = {
            QueuedJob.State.PENDING: "lightblue",
            QueuedJob.State.RUNNING: "yellow",
            QueuedJob.State.SUCCEEDED: "lightgreen",
            QueuedJob.State.FAILED: "lightpink",
        }

        # Add nodes (steps)
        for step_id, step in workflow.steps.items():
            color = colors.get(step.state, "lightgray")
            output.write(
                f'  "{step_id}" [label="{step_id}\\n({step.state})",'
                f" fillcolor={color}];\n"
            )

        # Add edges (dependencies)
        for step_id, step in workflow.steps.items():
            for dep in step.dependencies:
                output.write(f'  "{dep}" -> "{step_id}";\n')

        # Add workflow info
        output.write('  labelloc="t";\n')
        output.write(
            f'  label="Workflow: {workflow.name}\\nState: {workflow.state}";\n'
        )

        # Close the digraph
        output.write("}\n")

    async def cleanup(self, chancy: Chancy) -> int | None:
        async with chancy.pool.connection() as conn:
            async with conn.cursor() as cursor:
                await cursor.execute(
                    sql.SQL(
                        """
                        DELETE FROM {workflows}
                        WHERE state NOT IN ('pending', 'running')
                        AND ({rule})
                        """
                    ).format(
                        workflows=sql.Identifier(f"{chancy.prefix}workflows"),
                        rule=self.pruning_rule.to_sql(),
                    )
                )
                return cursor.rowcount

    def migrate_package(self) -> str:
        return "chancy.plugins.workflow.migrations"

    def migrate_key(self) -> str:
        return "workflow"

    def api_plugin(self) -> str | None:
        return "chancy.plugins.workflow.api.WorkflowApiPlugin"


class Sequence:
    """
    A simple sequential workflow.

    Sequences are a special case of workflows, where each step depends on the
    previous step. This forms a linear chain of jobs that are executed in
    order.

    Sequences are useful for defining sequences of jobs that must be executed
    in order, without the complexity of full workflows.

    Example
    -------

    .. code-block:: python
       :caption: example_sequence.py

        import asyncio
        from chancy import Job, Chancy
        from chancy.plugins.workflow import Sequence

        async def first():
            print("First")

        async def second():
            print("Second")

        async def third():
            print("Third")

        async def main():
            async with Chancy(dsn="postgresql://localhost/postgres") as chancy:
                sequence = Sequence("example", [
                    Job.from_func(first),
                    Job.from_func(second),
                    Job.from_func(third),
                ])
                await sequence.push(chancy)

        if __name__ == "__main__":
            asyncio.run(main())
    """

    def __init__(self, name: str, jobs: List[Job] = None):
        self.name = name
        self.jobs = jobs or []

    def add(self, job: Job) -> Self:
        """
        Add a job to the sequence.

        :param job: The job to add.
        """
        self.jobs.append(job)
        return self

    async def push(self, chancy: Chancy) -> str:
        """
        Push a new sequence to the database.

        :param chancy: The Chancy application.
        :return: The UUID of the newly created chain.
        """
        workflow = Workflow(self.name)
        for i, job in enumerate(self.jobs):
            step_id = f"step_{i}"
            dependencies = [f"step_{i - 1}"] if i > 0 else []
            workflow.add_step(step_id, job, dependencies)

        return await WorkflowPlugin.push(chancy, workflow)



import asyncio
import code

import click

from chancy import Chancy, Worker, Job, QueuedJob, Limit, Reference, Queue
from chancy.cli import run_async_command


@click.group(name="misc")
def misc_group():
    """
    Miscellaneous commands.
    """
    pass


@misc_group.command()
@click.option(
    "--to-version",
    "-t",
    help="The version to migrate to.",
    type=int,
)
@click.pass_context
@run_async_command
async def migrate(ctx: click.Context, to_version: int | None):
    """
    Migrate the database's core tables and any configured plugins up or down
    to the desired version.

    By default, migrates to the latest version, but can be controlled with
    the `--to-version` option.
    """
    chancy: Chancy = ctx.obj["app"]

    async with chancy:
        await chancy.migrate(to_version=to_version)


@misc_group.command()
@click.pass_context
@run_async_command
async def shell(ctx):
    """
    Start an interactive shell with the Chancy app instance and common
    imports already available.
    """
    chancy: Chancy = ctx.obj["app"]

    async with chancy:
        code.interact(
            local={
                "chancy": chancy,
                "Worker": Worker,
                "Job": Job,
                "QueuedJob": QueuedJob,
                "Limit": Limit,
                "Reference": Reference,
                "Queue": Queue,
                "asyncio": asyncio,
            }
        )



from typing import Callable

import click

from chancy import Chancy
from chancy.utils import import_string


@click.group()
@click.option(
    "--app",
    "-a",
    help="The import path for the Chancy app.",
    required=True,
)
@click.pass_context
def cli(ctx: click.Context, app: str):
    """
    Command line interface for Chancy.
    """
    ctx.ensure_object(dict)

    # Might be a Chancy instance or a function we need to call to get one.
    chancy_app: Chancy | Callable[[], Chancy] = import_string(app)

    if isinstance(chancy_app, Chancy):
        ctx.obj["app"] = chancy_app
    else:
        ctx.obj["app"] = chancy_app()


def main():
    from chancy.cli import misc, queue, worker

    cli.add_command(misc.misc_group)
    cli.add_command(queue.queue_group)
    cli.add_command(worker.worker_group)

    cli()


if __name__ == "__main__":
    main()



import asyncio
import functools
import multiprocessing
import os
from multiprocessing.context import BaseContext

import resource
import signal
import threading
from asyncio import Future
from concurrent.futures import ProcessPoolExecutor
from typing import Callable

from chancy.executors.base import Executor
from chancy.job import QueuedJob, Limit


class _TimeoutThread(threading.Thread):
    """
    A thread that will raise a TimeoutError after a specified number of
    seconds.
    """

    def __init__(self, timeout: int, cancel: threading.Event):
        super().__init__()
        self.timeout = timeout
        self.cancel = cancel

    def run(self):
        if self.cancel.wait(self.timeout) is True:
            # Our call to wait() returning True means that the flag was set
            # before the timeout elapsed, so we should cancel our alarm.
            return

        # If we reach this point, the timeout has elapsed, and we should raise
        # a TimeoutError back in the main process thread.
        os.kill(os.getpid(), signal.SIGALRM)


class ProcessExecutor(Executor):
    """
    .. note::

        Currently, this is the default executor used by Chancy.

    An Executor which uses a process pool to run its jobs.

    This executor is useful for running jobs that are CPU-bound, avoiding the
    GIL (Global Interpreter Lock) that Python uses to ensure thread safety.

    If a timeout is requested for a job, a separate thread will be spawned
    within each process to raise a TimeoutError if the job takes too long to
    complete.

    To use this executor, simply pass the import path to this class in the
    ``executor`` field of your queue configuration:

    .. code-block:: python

        async with Chancy(dsn="postgresql://localhost/postgres") as chancy:
            await chancy.declare(
                Queue(
                    name="default",
                    executor="chancy.executors.process.ProcessExecutor"
                )
            )

    :param queue: The queue that this executor is associated with.
    :param maximum_jobs_per_worker: The maximum number of jobs that each worker
                                    can run before being replaced.
    :param mp_context: The multiprocessing context to use. If not provided, the
                       default "spawn" context will be used, which is the
                       safest option on all platforms.
    """

    def __init__(
        self,
        worker,
        queue,
        *,
        maximum_jobs_per_worker: int = 100,
        mp_context: BaseContext | None = None,
    ):
        super().__init__(worker, queue)

        self.processes: dict[Future, QueuedJob] = {}
        self.pool = ProcessPoolExecutor(
            max_workers=queue.concurrency,
            max_tasks_per_child=maximum_jobs_per_worker,
            initializer=self.on_initialize_worker,
            # Becomes the default on Linux as of 3.14 due to potential crashes
            # in the `fork` method if the child process also works with threads.
            # We're using `spawn` explicitly to get ahead of the curve, however
            # this is slower than `fork`.
            mp_context=mp_context or multiprocessing.get_context("spawn"),
        )

    @staticmethod
    def on_initialize_worker():
        """
        This method is called in each worker process before it begins running
        jobs. It can be used to perform any necessary setup, such as loading
        NLTK datasets or calling ``django.setup()``.

        This isn't called once per job but once per worker process until
        :attr:`~ProcessExecutor.maximum_jobs_per_worker` is reached (if
        set). After that, the worker process is replaced with a new one.

        .. note::

            Care should be taken when overriding this method, as it is called
            within a separate process and may not have access to the same
            resources as the main process.
        """

    async def push(self, job: QueuedJob) -> Future:
        future: Future = self.pool.submit(self.job_wrapper, job)
        future.add_done_callback(
            functools.partial(
                self._on_job_completed, loop=asyncio.get_running_loop()
            )
        )
        self.processes[future] = job
        return future

    def __len__(self):
        return len(self.processes)

    @classmethod
    def job_wrapper(cls, job: QueuedJob):
        """
        This is the function that is actually started by the process pool
        executor. It's responsible for setting up necessary signals and limits,
        running the job, and returning the result.

        Subclasses can override this method to provide additional functionality
        or to change the way that jobs are run.

        .. note::

            Care should be taken when overriding this method, as it is called
            within a separate process and may not have access to the same
            resources as the main process.
        """
        func, kwargs = Executor.get_function_and_kwargs(job)
        if asyncio.iscoroutinefunction(func):
            raise ValueError(
                f"Function {job.func!r} is an async function, which is not"
                f" supported by the {cls.__name__!r} executor. Use the "
                "AsyncExecutor instead."
            )

        cleanup: list[Callable] = []
        for limit in job.limits:
            match limit.type_:
                case Limit.Type.TIME:
                    # We need this to work on platforms that don't support
                    # SIGALRM (looking at you, Windows). So we use a separate
                    # thread to raise a TimeoutError if the job takes too long.
                    signal.signal(signal.SIGALRM, cls.job_signal_handler)
                    cancel = threading.Event()
                    timeout_thread = _TimeoutThread(limit.value, cancel)
                    timeout_thread.start()
                    cleanup.append(
                        lambda: cancel.set() and timeout_thread.join()
                    )
                case Limit.Type.MEMORY:
                    previous_soft, _ = resource.getrlimit(resource.RLIMIT_AS)
                    resource.setrlimit(resource.RLIMIT_AS, (limit.value, -1))
                    cleanup.append(
                        lambda: resource.setrlimit(
                            resource.RLIMIT_AS, (previous_soft, -1)
                        )
                    )

        try:
            func(**kwargs)
        finally:
            for clean in cleanup:
                clean()

        return job

    @staticmethod
    def job_signal_handler(signum: int, frame):
        """
        Handles signals sent to a running job process.

        Subclasses can override this method to provide additional functionality
        or to change the way that signals are handled.

        .. note::

            Care should be taken when overriding this method, as it is called
            within a separate process and may not have access to the same
            resources as the main process.
        """
        # If we receive a SIGALRM signal, we should raise a TimeoutError, as
        # this is our _TimeoutThread telling us that the job has taken too long
        # to complete.
        if signum == signal.SIGALRM:
            raise TimeoutError("Job timed out.")

    def _on_job_completed(
        self, future: Future, loop: asyncio.AbstractEventLoop
    ):
        job = self.processes.pop(future)
        exc = future.exception()
        if exc is None:
            job = future.result()

        f = asyncio.run_coroutine_threadsafe(
            self.job_completed(job, exc),
            loop,
        )
        f.result()

    async def stop(self):
        """
        Stop the executor and clean up any resources.
        """
        # self.pool.shutdown(cancel_futures=True)
        await super().stop()



import asyncio
import inspect
from typing import Callable, Awaitable


class Hub:
    """
    An event hub for registering and emitting events.
    """

    def __init__(self):
        self._handlers: dict[str, list[Callable[..., Awaitable[None]]]] = {}
        self._waiters: dict[str, set[asyncio.Future]] = {}

    def on(self, event: str, f: Callable[..., Awaitable[None]]):
        """
        Register a handler for an event.
        """
        self._handlers.setdefault(event, []).append(f)

    def remove(self, event: str, handler: Callable[..., Awaitable[None]]):
        """
        Remove a handler from an event.
        """
        try:
            self._handlers.get(event, []).remove(handler)
        except ValueError:
            pass

    def clear(self):
        """
        Clear all handlers.
        """
        self._handlers = {}

        for event in self._waiters:
            for waiter in self._waiters[event]:
                if not waiter.done():
                    waiter.cancel()

        self._waiters = {}

    async def emit(self, event: str, body):
        """
        Emit an event with the given body.
        """
        for handler in self._handlers.get(event, []):
            # Check if the handler is async
            if inspect.iscoroutinefunction(handler):
                await handler(body)
            else:
                handler(body)

        # Emit wildcard handlers.
        if "*" in self._handlers:
            for handler in self._handlers["*"]:
                if inspect.iscoroutinefunction(handler):
                    await handler(event, body)
                else:
                    handler(event, body)

        # Resolve any waiting futures
        waiters = self._waiters.get(event, [])
        for waiter in waiters:
            if not waiter.done():
                waiter.set_result(body)
        self._waiters[event] = {w for w in waiters if not w.done()}

    async def wait_for(self, event: str, timeout: float | None = None):
        """
        Wait for a specific event to occur.

        :param event: The event to wait for.
        :param timeout: The maximum time to wait for the event (in seconds).
        """
        future = asyncio.get_running_loop().create_future()
        future.add_done_callback(
            lambda _: self._waiters.get(event, set()).discard(future)
        )
        self._waiters.setdefault(event, set()).add(future)

        done, pending = await asyncio.wait(
            [future],
            timeout=timeout,
            return_when=asyncio.FIRST_COMPLETED,
        )

        for task in pending:
            task.cancel()

        return



__all__ = (
    "Chancy",
    "Worker",
    "Queue",
    "Job",
    "QueuedJob",
    "Limit",
    "Reference",
)

from chancy.app import Chancy
from chancy.queue import Queue
from chancy.worker import Worker
from chancy.job import Limit, Job, QueuedJob, Reference



from datetime import datetime, timezone

import pytest
from psycopg import sql

from chancy import Chancy, Job, Queue, Worker
from chancy.utils import chancy_uuid, timed_block


def dummy_job():
    pass


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "chancy",
    [
        {
            "notifications": False,
        }
    ],
    indirect=True,
)
async def test_busy_chancy(chancy: Chancy, worker_no_start: Worker):
    """
    Tests the performance of the worker when the Chancy instance has 10 million
    completed jobs and 40,000 pending jobs.

    This is NOT testing optimal cases - it's here to catch egregious performance
    regressions. For example our first test checks job fetches are under 100ms,
    but it's typically sub 1ms in reality.
    """
    queue = Queue("default", concurrency=10)
    await chancy.declare(queue)

    async with chancy.pool.connection() as conn:
        async with conn.cursor() as cursor:
            async with cursor.copy(
                sql.SQL(
                    "COPY {jobs_table} ("
                    "   id,"
                    "   queue,"
                    "   state,"
                    "   func,"
                    "   kwargs,"
                    "   meta,"
                    "   created_at,"
                    "   scheduled_at"
                    ")"
                    " FROM STDIN"
                ).format(jobs_table=sql.Identifier(f"{chancy.prefix}jobs"))
            ) as copy:
                for i in range(10_000_000):
                    await copy.write_row(
                        (
                            chancy_uuid(),
                            "default",
                            "succeeded",
                            "dummy_job",
                            "{}",
                            "{}",
                            datetime.now(timezone.utc),
                            datetime.now(timezone.utc),
                        )
                    )

                j = Job.from_func(dummy_job)

                for i in range(40_000):
                    await copy.write_row(
                        (
                            chancy_uuid(),
                            "default",
                            "pending",
                            j.func,
                            "{}",
                            "{}",
                            datetime.now(timezone.utc),
                            datetime.now(timezone.utc),
                        )
                    )

            await conn.commit()

        with timed_block() as timer:
            await worker_no_start.fetch_jobs(queue, conn, up_to=1)
        assert timer.elapsed < 0.1

        with timed_block() as timer:
            await worker_no_start.fetch_jobs(queue, conn, up_to=100)
        assert timer.elapsed < 0.1



import asyncio
import functools
import logging
from typing import Any, Iterator
from functools import cached_property, cache

from psycopg import sql, Cursor, AsyncCursor
from psycopg.rows import dict_row
from psycopg.types.json import Json
from psycopg_pool import AsyncConnectionPool, ConnectionPool

from chancy.migrate import Migrator
from chancy.queue import Queue
from chancy.job import Reference, Job, QueuedJob
from chancy.plugin import Plugin
from chancy.utils import chancy_uuid, chunked, json_dumps


@cache
def _setup_default_logger():
    logger = logging.getLogger("chancy")
    logger.setLevel(logging.DEBUG)

    handler = logging.StreamHandler()
    handler.setFormatter(
        logging.Formatter(
            fmt="%(asctime)s • %(levelname)s • %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
        )
    )

    logger.addHandler(handler)
    return logger


class Chancy:
    """
    A Chancy application, along with all of its configuration and common
    functionality.

    Chancy is asyncio-first, and all of its methods are async by default:

    .. code-block:: python

        import asyncio
        from chancy import Chancy, Job, Queue

        def my_job():
            print("Hello, world!")

        async def main():
            async with Chancy("postgresql://localhost/chancy") as chancy:
                await chancy.migrate()
                await chancy.declare(Queue("default"))
                await chancy.push(Job.from_func(my_job))

        asyncio.run(main())

    Chancy does also provide a minimal synchronous interface for pushing jobs
    onto the queue for codebases that are not using asyncio:

    .. code-block:: python

        from chancy import Chancy, Job, Queue

        def my_job():
            print("Hello, world!")

        with Chancy("postgresql://localhost/chancy") as chancy:
            chancy.sync_push(Job.from_func(my_job))

    And of course, it's just Postgres under the hood, so you can always just
    insert jobs directly into the database if you're feeling brave.

    :param dsn: The DSN to connect to the database.
    :param plugins: The plugins to use with the application.
    :param prefix: A prefix appended to all table names, which can be used to
        namespace the tables for multiple applications or tenants.
    :param min_connection_pool_size: The minimum number of connections to keep
        in the pool.
    :param max_connection_pool_size: The maximum number of connections to keep
        in the pool.
    :param poll_reconnect_timeout: The number of seconds to wait before
        attempting to reconnect to the database if a connection is lost.
    :param notifications: Enables or disables emitting notifications using
        postgres's NOTIFY/LISTEN feature.
    :param log: The logger to use for all application logging.
    """

    @staticmethod
    def _ensure_pool_is_open(f):
        """
        A decorator which ensures the connection pool is open before calling
        the wrapped function.

        It's never ideal to rely on this, since it will never _close_ the pool,
        but users expect to be able to just call push() and have it work.
        """

        @functools.wraps(f)
        async def wrapper(self, *args, **kwargs):
            await self.pool.open()
            return await f(self, *args, **kwargs)

        return wrapper

    @staticmethod
    def _ensure_sync_pool_is_open(f):
        """
        A decorator which ensures the connection pool is open before calling
        the wrapped function.

        It's never ideal to rely on this, since it will never _close_ the pool,
        but users expect to be able to just call push() and have it work.
        """

        @functools.wraps(f)
        def wrapper(self, *args, **kwargs):
            self.sync_pool.open()
            return f(self, *args, **kwargs)

        return wrapper

    @staticmethod
    def _ensure_pool_is_open_async_iter(f):
        """
        A decorator which ensures the connection pool is open before calling
        the wrapped function.

        It's never ideal to rely on this, since it will never _close_ the pool,
        but users expect to be able to just call push() and have it work.
        """

        @functools.wraps(f)
        async def wrapper(self, *args, **kwargs):
            await self.pool.open()
            async for record in f(self, *args, **kwargs):
                yield record

        return wrapper

    def __init__(
        self,
        dsn: str,
        *,
        plugins: list[Plugin] = None,
        prefix: str = "chancy_",
        min_connection_pool_size: int = 1,
        max_connection_pool_size: int = 10,
        poll_reconnect_timeout: int = 60 * 5,
        notifications: bool = True,
        log: logging.Logger | None = None,
    ):
        #: The DSN to connect to the database.
        self.dsn = dsn
        #: The plugins to use with the application.
        self.plugins = plugins or []
        #: A prefix appended to all table names, which can be used to
        #: namespace the tables for multiple applications or tenants.
        self.prefix = prefix
        #: The minimum number of connections to keep in the pool.
        self.min_connection_pool_size = min_connection_pool_size
        #: The maximum number of connections to keep in the pool.
        self.max_connection_pool_size = max_connection_pool_size
        #: The number of seconds to wait before attempting to reconnect to the
        #: database if a connection is lost.
        self.poll_reconnect_timeout = poll_reconnect_timeout
        #: Enables or disables emitting notifications using postgres's
        #: NOTIFY/LISTEN feature.
        self.notifications = notifications
        #: The logger to use for all application logging.
        self.log = log or _setup_default_logger()

    async def __aenter__(self):
        await self.pool.open()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.pool.close()

    def __enter__(self):
        self.sync_pool.open()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.sync_pool.close()
        return False

    @cached_property
    def pool(self) -> AsyncConnectionPool:
        """
        The asyncio connection pool used to interact with the database.
        """
        return AsyncConnectionPool(
            self.dsn,
            min_size=self.min_connection_pool_size,
            max_size=self.max_connection_pool_size,
            reconnect_timeout=self.poll_reconnect_timeout,
            # Opening the pool here is deprecated, and should be done
            # explicitly by the user with a context manager or open/close.
            open=False,
        )

    @cached_property
    def sync_pool(self) -> ConnectionPool:
        """
        The synchronous connection pool used to interact with the database.
        """
        return ConnectionPool(
            self.dsn,
            min_size=self.min_connection_pool_size,
            max_size=self.max_connection_pool_size,
            reconnect_timeout=self.poll_reconnect_timeout,
        )

    @_ensure_pool_is_open
    async def migrate(self, *, to_version: int | None = None):
        """
        Migrate the database to the latest version.

        This will migrate the core database tables as well as any enabled
        plugins. If a version is provided, the database will be migrated up
        _or_ down to that version as necessary.

        .. code-block:: python

            async with Chancy("postgresql://localhost/chancy") as chancy:
                await chancy.migrate()

        Can also be run with the CLI:

        .. code-block:: bash

            chancy --app worker.chancy misc migrate

        .. warning::

            You should never run migrations in production without first testing
            them in a development environment. Migrations can be destructive and
            may cause data loss if not written carefully, or take a long time to
            complete if the database is large.

        :param to_version: The version to migrate to. If not provided, the
            database will be migrated to the latest version.
        """
        migrator = Migrator("chancy", "chancy.migrations", prefix=self.prefix)
        async with self.pool.connection() as conn:
            await migrator.migrate(conn, to_version=to_version)

        for plugin in self.plugins:
            await plugin.migrate(self, to_version=to_version)

    @_ensure_pool_is_open
    async def declare(self, queue: Queue, *, upsert: bool = False) -> Queue:
        """
        Declare a queue in the database.

        Queues in Chancy are the primary mechanism for organizing and
        executing jobs. They can be configured with various options to
        control how jobs are processed, including the concurrency level,
        rate limiting, and tags to restrict which workers can process jobs
        from the queue.

        Queues exist globally, configured at runtime via the database and
        can be updated, reassigned, or deleted at any time without requiring
        a restart of the workers.

        This will create a new queue in the database with the provided
        configuration. If the queue already exists, no changes will be
        made unless the `upsert` parameter is also set to `True`.

        .. code-block:: python

            async with Chancy("postgresql://localhost/chancy") as chancy:
                await chancy.declare(Queue("default"))

        :param queue: The queue to declare.
        :param upsert: If `True`, the queue will be updated if it already
            exists. Defaults to `False`.
        :return: The queue as it exists in the database.
        """
        async with self.pool.connection() as conn:
            async with conn.cursor() as cursor:
                queue = await self.declare_ex(cursor, queue, upsert=upsert)
                await self.notify(cursor, "queue.declared", {"q": queue.name})
                return queue

    async def declare_ex(
        self, cursor: AsyncCursor, queue: Queue, *, upsert: bool = False
    ) -> Queue:
        """
        Declare a queue in the database using a specific cursor.

        This is a low-level method that allows for more control over the
        database connection and transaction management. It is recommended to
        use the higher-level `declare` method in most cases.

        :param cursor: The cursor to use for the operation.
        :param queue: The queue to declare.
        :param upsert: If `True`, the queue will be updated if it already
            exists. Defaults to `False`.
        """
        await cursor.execute(
            self._declare_sql(upsert),
            {
                "name": queue.name,
                "state": queue.state.value,
                "concurrency": queue.concurrency,
                "tags": list(queue.tags),
                "executor": queue.executor,
                "executor_options": Json(queue.executor_options),
                "polling_interval": queue.polling_interval,
                "rate_limit": queue.rate_limit,
                "rate_limit_window": queue.rate_limit_window,
            },
        )

        result = await cursor.fetchone()
        return Queue(
            name=queue.name,
            state=result[0],
            concurrency=result[1],
            tags=set(result[2]),
            polling_interval=result[3],
            executor=result[4],
            executor_options=result[5],
            rate_limit=result[6],
            rate_limit_window=result[7],
        )

    @_ensure_pool_is_open
    async def push(self, job: Job) -> Reference:
        """
        Push a job onto the queue.

        This method will push a job onto the queue, making it available for
        processing by workers. A :class:`Reference` object is returned that
        can be used to track the progress of the job and retrieve the result
        when it is complete.

        .. code-block:: python

            async with Chancy("postgresql://localhost/chancy") as chancy:
                await chancy.push(Job.from_func(my_job))

        .. seealso::

            :meth:`sync_push` for a synchronous version of this method.

        :param job: The job to push onto the queue.
        :return: A reference to the job in the queue.
        """
        async with self.pool.connection() as conn:
            async with conn.cursor() as cursor:
                return (await self.push_many_ex(cursor, [job]))[0]

    @_ensure_sync_pool_is_open
    def sync_push(self, job: Job) -> Reference:
        """
        Synchronously push a job onto the queue.

        This method provides a synchronous interface to push a job onto the
        queue.

        .. code-block:: python

            with Chancy("postgresql://localhost/chancy") as chancy:
                chancy.sync_push(Job.from_func(my_job))

        .. seealso::

            :meth:`push` for an asynchronous version of this method.

        :param job: The job to push onto the queue.
        :return: A reference to the job in the queue.
        """
        with self.sync_pool.connection() as conn:
            with conn.cursor() as cursor:
                return self.sync_push_many_ex(cursor, [job])[0]

    @_ensure_pool_is_open_async_iter
    async def push_many(
        self, jobs: list[Job], *, batch_size: int = 1000
    ) -> Iterator[list[Reference]]:
        """
        Push multiple jobs onto the queue.

        This method will push multiple jobs onto the queue, making them
        available for processing by workers. Jobs are committed in batches,
        with the size of each batch controlled by the `batch_size` parameter.
        Each batch is committed within its own transaction to limit peak
        memory usage.

        Returns an iterator of lists of references to the jobs in the queue,
        one list per batch.

        .. seealso::

            :meth:`sync_push_many` for a synchronous version of this method.

        :param jobs: The jobs to push onto the queue.
        :param batch_size: The number of jobs to push in each batch. Defaults
            to 1000.
        :return: An iterator of lists of references to the jobs in the queue.
        """
        async with self.pool.connection() as conn:
            async with conn.cursor() as cursor:
                for chunk in chunked(jobs, batch_size):
                    async with conn.transaction():
                        yield await self.push_many_ex(cursor, chunk)

    @_ensure_sync_pool_is_open
    def sync_push_many(
        self, jobs: list[Job], *, batch_size: int = 1000
    ) -> Iterator[list[Reference]]:
        """
        Synchronously push multiple jobs onto the queue.

        This method provides a synchronous interface to push multiple jobs onto
        the queue.

        Returns an iterator of lists of references to the jobs in the queue,
        one list per batch.

        .. seealso::

            :meth:`push_many` for an asynchronous version of this method.

        :param jobs: The jobs to push onto the queue.
        :param batch_size: The number of jobs to push in each batch. Defaults
            to 1000.
        :return: An iterator of lists of references to the jobs in the queue.
        """
        with self.sync_pool.connection() as conn:
            with conn.cursor() as cursor:
                for chunk in chunked(jobs, batch_size):
                    with conn.transaction():
                        yield self.sync_push_many_ex(cursor, chunk)

    async def push_many_ex(
        self, cursor: AsyncCursor, jobs: list[Job]
    ) -> list[Reference]:
        """
        Push multiple jobs onto the queue using a specific cursor.

        This is a low-level method that allows for more control over the
        database connection and transaction management. It is recommended to
        use the higher-level `push_many` method in most cases.

        The main advantage of using this function is to ensure your jobs are
        only pushed if some other operation is successful. For example, you
        only want to send an email confirmation if account creation actually
        succeeds. For example:

        .. seealso::

            :meth:`sync_push_many_ex` for a synchronous version of this method.

        :param cursor: The cursor to use for the operation.
        :param jobs: The jobs to push onto the queue.
        :return: A list of references to the jobs in the queue.
        """
        references = []
        for job in jobs:
            await cursor.execute(
                self._push_job_sql(),
                self._get_job_params(job),
            )
            record = await cursor.fetchone()
            references.append(Reference(record[0]))

        if self.notifications:
            for queue in set(job.queue for job in jobs):
                await self.notify(cursor, "queue.pushed", {"q": queue})

        return references

    def sync_push_many_ex(
        self, cursor: Cursor, jobs: list[Job]
    ) -> list[Reference]:
        """
        Synchronously push multiple jobs onto the queue using a specific cursor.

        This is a low-level method that allows for more control over the
        database connection and transaction management. It is recommended to
        use the higher-level `sync_push_many` method in most cases.

        .. seealso::

            :meth:`push_many_ex` for an asynchronous version of this method.

        :param cursor: The cursor to use for the operation.
        :param jobs: The jobs to push onto the queue.
        :return: A list of references to the jobs in the queue.
        """
        references = []
        for job in jobs:
            cursor.execute(
                self._push_job_sql(),
                self._get_job_params(job),
            )
            record = cursor.fetchone()
            references.append(Reference(record[0]))

        for queue in set(job.queue for job in jobs):
            self.sync_notify(cursor, "queue.pushed", {"q": queue})

        return references

    @_ensure_pool_is_open
    async def get_job(self, ref: Reference) -> QueuedJob | None:
        """
        Resolve a reference to a job instance.

        If the job no longer exists, returns ``None``.

        :param ref: The reference to the job to retrieve.
        """
        async with self.pool.connection() as conn:
            async with conn.cursor(row_factory=dict_row) as cursor:
                await cursor.execute(self._get_job_sql(), [ref.identifier])
                record = await cursor.fetchone()
                if record is None:
                    raise KeyError(f"Job {ref.identifier} not found.")
                return QueuedJob.unpack(record)

    async def wait_for_job(
        self,
        ref: Reference,
        *,
        interval: int = 1,
        timeout: float | int | None = None,
    ) -> QueuedJob | None:
        """
        Wait for a job to complete.

        This method will loop until the job referenced by the provided
        reference has completed. The interval parameter controls how often
        the job status is checked. This will not block the event loop, so
        other tasks can run while waiting for the job to complete.

        If the job no longer exists, returns ``None``.

        :param ref: The reference to the job to wait for.
        :param interval: The number of seconds to wait between checks.
        :param timeout: The maximum number of seconds to wait for the job to
            complete. If not provided, the method will wait indefinitely.
        """
        async with asyncio.timeout(timeout):
            while True:
                job = await self.get_job(ref)
                if job is None or job.state in {
                    QueuedJob.State.SUCCEEDED,
                    QueuedJob.State.FAILED,
                }:
                    return job
                await asyncio.sleep(interval)

    @_ensure_pool_is_open
    async def get_all_queues(self) -> list[Queue]:
        """
        Get all queues known to the cluster, regardless of their status
        or if they're assigned to any workers.
        """
        async with self.pool.connection() as conn:
            async with conn.cursor(row_factory=dict_row) as cursor:
                await cursor.execute(self._get_all_queues_sql())
                return [Queue.unpack(record) async for record in cursor]

    @_ensure_pool_is_open
    async def get_queue(self, name: str) -> Queue:
        """
        Get a specific queue by name.

        :param name: The name of the queue to retrieve.
        """
        async with self.pool.connection() as conn:
            async with conn.cursor(row_factory=dict_row) as cursor:
                await cursor.execute(self._get_queue_sql(), [name])
                record = await cursor.fetchone()
                if record is None:
                    raise KeyError(f"Queue {name!r} not found.")
                return Queue.unpack(record)

    async def delete_queue(self, name: str, *, purge_jobs: bool = True):
        """
        Delete a queue by name.

        By default, also deletes all jobs in the queue. If you want to keep
        the jobs, set `purge_jobs` to `False`.

        :param name: The name of the queue to delete.
        :param purge_jobs: If `True`, all jobs in the queue will be deleted
            along with the queue. If `False`, the jobs will be left in the
            database.
        """
        async with self.pool.connection() as conn:
            async with conn.cursor() as cursor:
                await cursor.execute(
                    sql.SQL(
                        """
                        DELETE FROM {queues}
                        WHERE name = %s
                        """
                    ).format(queues=sql.Identifier(f"{self.prefix}queues")),
                    [name],
                )
                if purge_jobs:
                    await cursor.execute(
                        sql.SQL(
                            """
                            DELETE FROM {jobs}
                            WHERE queue = %s
                            """
                        ).format(jobs=sql.Identifier(f"{self.prefix}jobs")),
                        [name],
                    )

    @_ensure_pool_is_open
    async def get_all_workers(self) -> list[dict[str, Any]]:
        """
        Get all workers known to the cluster, regardless of their status.
        """
        async with self.pool.connection() as conn:
            async with conn.cursor(row_factory=dict_row) as cursor:
                await cursor.execute(self._get_all_workers_sql())
                return [record async for record in cursor]

    async def notify(
        self, cursor: AsyncCursor, event: str, payload: dict[str, Any]
    ):
        """
        Send a notification via Postgres NOTIFY/LISTEN to all other listening
        workers.

        If notifications are disabled on the associated Chancy app, this
        method will do nothing.

        .. seealso::

            :meth:`sync_notify` for a synchronous version of this method.
        """
        if not self.notifications:
            return

        await cursor.execute(
            "SELECT pg_notify(%s, %s)",
            [
                f"{self.prefix}events",
                json_dumps({"t": event, **payload}),
            ],
        )

    def sync_notify(self, cursor: Cursor, event: str, payload: dict[str, Any]):
        """
        Send a notification via Postgres NOTIFY/LISTEN to all other listening
        workers.

        If notifications are disabled on the associated Chancy app, this
        method will do nothing.

        .. seealso::

            :meth:`notify` for an asynchronous version of this method.
        """
        if not self.notifications:
            return

        cursor.execute(
            "SELECT pg_notify(%s, %s)",
            [
                f"{self.prefix}events",
                json_dumps({"t": event, **payload}),
            ],
        )

    def _get_all_workers_sql(self):
        return sql.SQL(
            """
            SELECT
                *
            FROM
                {workers}
            """
        ).format(workers=sql.Identifier(f"{self.prefix}workers"))

    def _get_queue_sql(self):
        return sql.SQL(
            """
            SELECT
                *
            FROM
                {queues}
            WHERE
                name = %s
            """
        ).format(queues=sql.Identifier(f"{self.prefix}queues"))

    def _get_all_queues_sql(self):
        return sql.SQL(
            """
            SELECT
                *
            FROM
                {queues}
            """
        ).format(queues=sql.Identifier(f"{self.prefix}queues"))

    def _push_job_sql(self):
        return sql.SQL(
            """
            INSERT INTO
                {jobs} (
                    id,
                    queue,
                    func,
                    kwargs,
                    limits,
                    meta,
                    priority,
                    max_attempts,
                    scheduled_at,
                    unique_key
                )
            VALUES (
                %(id)s,
                %(queue)s,
                %(func)s,
                %(kwargs)s,
                %(limits)s,
                %(meta)s,
                %(priority)s,
                %(max_attempts)s,
                %(scheduled_at)s,
                %(unique_key)s
            )
            ON CONFLICT (unique_key)
            WHERE
                unique_key IS NOT NULL
                    AND state NOT IN ('succeeded', 'failed')
            DO UPDATE
               SET
                   state = EXCLUDED.state
            RETURNING id;
            """
        ).format(jobs=sql.Identifier(f"{self.prefix}jobs"))

    def _get_job_sql(self):
        return sql.SQL(
            """
            SELECT
                *
            FROM
                {jobs}
            WHERE
                id = %s
            """
        ).format(jobs=sql.Identifier(f"{self.prefix}jobs"))

    def _declare_sql(self, upsert: bool):
        action = sql.SQL(
            """
            UPDATE SET
                name = EXCLUDED.name
            """
        )
        if upsert:
            action = sql.SQL(
                """
                UPDATE SET
                    state = EXCLUDED.state,
                    concurrency = EXCLUDED.concurrency,
                    tags = EXCLUDED.tags,
                    executor = EXCLUDED.executor,
                    executor_options = EXCLUDED.executor_options,
                    polling_interval = EXCLUDED.polling_interval,
                    rate_limit = EXCLUDED.rate_limit,
                    rate_limit_window = EXCLUDED.rate_limit_window
                """
            )

        return sql.SQL(
            """
            INSERT INTO {queues} (
                name,
                state,
                concurrency,
                tags,
                executor,
                executor_options,
                polling_interval,
                rate_limit,
                rate_limit_window
            ) VALUES (
                %(name)s,
                %(state)s,
                %(concurrency)s,
                %(tags)s,
                %(executor)s,
                %(executor_options)s,
                %(polling_interval)s,
                %(rate_limit)s,
                %(rate_limit_window)s
            )
            ON CONFLICT (name) DO
                {action}
            RETURNING 
                state,
                concurrency,
                tags,
                polling_interval,
                executor,
                executor_options,
                rate_limit,
                rate_limit_window;
            """
        ).format(
            queues=sql.Identifier(f"{self.prefix}queues"),
            action=action,
        )

    @staticmethod
    def _get_job_params(job: Job) -> dict:
        """
        Get the parameters for a job to be inserted into the database.

        :param job: The job to get parameters for.
        :return: A dictionary of parameters for the job.
        """
        return {
            "id": chancy_uuid(),
            "queue": job.queue,
            "func": job.func,
            "kwargs": Json(job.kwargs or {}),
            "limits": Json([limit.serialize() for limit in job.limits]),
            "meta": Json(job.meta),
            "priority": job.priority,
            "max_attempts": job.max_attempts,
            "scheduled_at": job.scheduled_at,
            "unique_key": job.unique_key,
        }
